/*  most parts of this code ripped from whois 4.5.7 by Marco d'Itri
    rest (c) 2001 Brad Spengler   */

#include "includes.h"
#include "data.h"
#include "whois.h"
#include "snorthead.h"
#include <errno.h>

int sockfd;
char domainname[256];
char *wee1;
char qstring[260] = "\0";

int nslookup(char *host,int ip){
        struct hostent *he;
        struct in_addr a;
	int i,match = 0;	
	char ipaddy[16];

	if(!ip) {
		if((he = gethostbyname(host)) != NULL){
		printf("<br><br><b>%s resolves to:<br>",host);
		for(i=0;he->h_addr_list[i];i++) {
			grzero(&ipaddy,sizeof(ipaddy));
			snprintf(ipaddy,16,"%u.%u.%u.%u",
			(unsigned char)he->h_addr_list[i][0],
			(unsigned char)he->h_addr_list[i][1],
			(unsigned char)he->h_addr_list[i][2],
			(unsigned char)he->h_addr_list[i][3]);
			printf("%s<br>",ipaddy);
			if(!strcmp(wee1,ipaddy))
				match = 1;
		}
		if (!match)
			printf("<BR>DNS forward doesn't match DNS reverse.<BR>"
			"Trust the IP information, not the DNS.</b>");
		else
			printf("<BR>DNS forward matches DNS reverse.<BR>"
			"DNS and IP information can be trusted.</b>");
		}
	} else {
	grzero(&domainname,sizeof(domainname));
        if(!inet_aton(host,&a)){
                printf("Can't resolve hostname from the IP.</b>");
                return 1;
        }
        if((he = gethostbyaddr(&a,sizeof(a),AF_INET)) == NULL) {
                printf("Can't resolve hostname from the IP.</b>");
                return 1;
        }

	printf("<b>%s resolves to %s.",host,he->h_name);			
	strncpy(domainname,he->h_name,sizeof(domainname)-1);
	}
	return 0;
}

/* check if dom ends with an element of tldlist[] */
int domfind(const char *dom, const char *tldlist[])
{
    int i;

    for (i = 0; tldlist[i]; i++)
        if (domcmp(dom, tldlist[i]))
            return 1;
    return 0;
}

/* check if dom ends with tld */
int domcmp(const char *dom, const char *tld)
{
    const char *p, *q;

    for (p = dom; *p != '\0'; p++); p--;        /* move to the last char */
    for (q = tld; *q != '\0'; q++); q--;
    while (p >= dom && q >= tld && tolower(*p) == *q) { /* compare backwards */
        if (q == tld)                           /* start of the second word? */
            return 1;
        p--; q--;
    }
    return 0;
}


unsigned long myinet_aton(const char *s)
{
    int a, b, c, d;

    if (!s)
        return 0;
    if (sscanf(s, "%d.%d.%d.%d", &a, &b, &c, &d) != 4)
        return 0;
    return (a << 24) + (b << 16) + (c << 8) + d;
}

const char *query_crsnic(const int sock, const char *query)
{
    char *temp, buf[100], *ret = NULL;
    FILE *fi;

    temp = malloc(strlen(query) + 1 + 2 + 1);
    *temp = '=';
    strcpy(temp + 1, query);
    strcat(temp, "\r\n");

    fi = fdopen(sock, "r");
    if (write(sock, temp, strlen(temp)) < 0)
        err_sys("write");
    while (fgets(buf, 100, fi)) {
        /* If there are multiple matches only the server of the first record
           is queried */
        if (strncmp(buf, "   Whois Server:", 16) == 0 && !ret) {
            char *p, *q;

            for (p = buf; *p != ':'; p++);      /* skip until colon */
            for (p++; *p == ' '; p++);          /* skip colon and spaces */
            ret = malloc(strlen(p) + 1);
            for (q = ret; *p != '\n' && *p != '\r'; *q++ = *p++); /*copy data*/
            *q = '\0';
        }
        fputs(buf, stdout);
    }
    if (ferror(fi))
        err_sys("fgets");

    free(temp);
    return ret;
}

int openconn(const char *server, const char *port)
{
    int fd;
    struct hostent *hostinfo;
    struct servent *servinfo;
    struct sockaddr_in saddr;

    if ((hostinfo = gethostbyname(server)) == NULL)
        err_quit(("Host %s not found."), server);
    if ((fd = socket(PF_INET, SOCK_STREAM, IPPROTO_IP)) < 0)
        err_sys("socket");
    grzero(&saddr, sizeof(saddr));
    saddr.sin_addr = *(struct in_addr *) hostinfo->h_addr;
    saddr.sin_family = AF_INET;
    if (!port) {
        saddr.sin_port = htons(43);
    } else if ((saddr.sin_port = htons(atoi(port))) == 0) {
        if ((servinfo = getservbyname(port, "tcp")) == NULL)
            err_quit(("%s/tcp: unknown service"), port);
        saddr.sin_port = servinfo->s_port;
    }
    if (connect(fd,(struct sockaddr *)&saddr, sizeof(saddr)) < 0)
        err_sys("connect");
    return (fd);
}

void closeconn(const int fd)
{
    close(fd);
}

/* Error routines */
void err_sys(const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, ": %s\n", strerror(errno));
    va_end(ap);
    exit(2);
}

void err_quit(const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    fputs("\n", stderr);
    va_end(ap);
    exit(2);
}

const char *whichwhois(const char *s)
{
    unsigned long ip;
    unsigned int i;

    /* -v or -t has been used */
    if (*s == '\0')
        return "whois.ripe.net";

    /* email address */
    if (grchr(s, '@',strlen(s)))
        return "";

    /* no dot and no hyphen means it's a NSI NIC handle or ASN (?) */
    if (!strpbrk(s, ".-")) {
        const char *p;

        for (p = s; *p != '\0'; p++);           /* go to the end of s */
        if (strncasecmp(p - 2, "jp", 2) == 0) /* JP NIC handle */
            return "whois.nic.ad.jp";
        if (*p == '!')  /* NSI NIC handle */
            return "whois.networksolutions.com";
        else /* it's a NSI NIC handle or something we don't know about */
            return "";
    }
    /* smells like an IP? */
    if ((ip = myinet_aton(s))) {
        for (i = 0; ip_assign[i].serv; i++)
            if ((ip & ip_assign[i].mask) == ip_assign[i].net)
                return ip_assign[i].serv;
        return "whois.arin.net";
    }

    /* check TLD list */
    for (i = 0; tld_serv[i]; i += 2)
        if (domcmp(s, tld_serv[i]))
            return tld_serv[i + 1];

    /* no dot but hyphen */
    if (!grchr(s, '.',strlen(s))) {
        /* search for strings at the start of the word */
        for (i = 0; nic_handles[i]; i += 2)
            if (!strncasecmp(s, nic_handles[i], strlen(nic_handles[i])))
                return nic_handles[i + 1];
        return "whois.arin.net";
    }

    /* has dot and hypen and it's not in tld_serv[], WTF is it? */
	return "";
}

char *queryformat(const char *server, const char *flags, const char 
*query)
{
    char *buf;
    int i, isripe = 0;

    /* +10 for CORE; +2 for \r\n; +1 for NULL */
    buf = malloc(strlen(flags) + strlen(query) + 10 + 2 + 1);
    *buf = '\0';
    for (i = 0; ripe_servers[i]; i++)
        if (strcmp(server, ripe_servers[i]) == 0) {
            strcat(buf, "-V " IDSTRING " ");
            isripe = 1;
            break;
        }
    if (!isripe)
        for (i = 0; ripe_servers_old[i]; i++)
            if (strcmp(server, ripe_servers_old[i]) == 0) {
                strcat(buf, "-V" IDSTRING " ");
                isripe = 1;
                break;
            }
    if (*flags != '\0') {
        if (!isripe && strcmp(server, "whois.corenic.net") != 0)
            puts(("Warning: RIPE flags ignored for a traditional 
server."));
        else
            strcat(buf, flags);
    }
    if (!isripe &&
            (strcmp(server, "whois.arin.net") == 0 ||
             strcmp(server, "whois.nic.mil") == 0) &&
            strncasecmp(query, "AS", 2) == 0 &&
            query[2] >= '0' && query[2] <= '9')
        sprintf(buf, "AS %s", query + 2);       /* fix query for ARIN */
    else if (!isripe && strcmp(server, "whois.corenic.net") == 0)
        sprintf(buf, "--machine %s", query);    /* machine readable output 
*/
    else if (!isripe && strcmp(server, "whois.ncst.ernet.in") == 0 &&
             !grchr(query, ' ',strlen(query)))
        sprintf(buf, "domain %s", query);       /* ask for a domain */
    else if (!isripe && strcmp(server, "whois.nic.ad.jp") == 0) {
        char *lang = getenv("LANG");    /* not a perfect check, but... */
        if (!lang || (strncmp(lang, "ja", 2) != 0))
            sprintf(buf, "%s/e", query);        /* ask for english text */
        else
            strcat(buf, query);
    } else
        strcat(buf, query);
    return buf;
}

void do_query(const int sock, const char *query)
{
    char buf[200], *p;
    FILE *fi;
    int i = 0, hide = 0;

    fi = fdopen(sock, "r");
    if (write(sock, query, strlen(query)) < 0)
        err_sys("write");
    while (fgets(buf, 200, fi)) {       /* XXX errors? */
        if (hide == 1) {
            if (strncmp(buf, hide_strings[i+1], strlen(hide_strings[i+1]))==0)
                hide = 2;       /* stop hiding */
            continue;           /* hide this line */
        }
        if (hide == 0) {
            for (i = 0; hide_strings[i] != NULL; i += 2) {
                if (strncmp(buf, hide_strings[i], strlen(hide_strings[i]))==0){
                    hide = 1;   /* start hiding */
                    break;
                }
            }
            if (hide == 1)
                continue;       /* hide the first line */
        }
        for (p = buf; *p && *p != '\r' && *p != '\n'; p++);
        *p = '\0';
        fprintf(stdout, "%s\n", buf);
    }
    if (ferror(fi))
        err_sys("fgets");

}

int whois(char *wee) {
	char fstring[64] = "\0";
	char *p;
	const char *port = NULL;
	extern char *optarg;
	extern int optind;
	const char *server = NULL;
	
	if (!server && domfind(wee,gtlds)) {
		sockfd = openconn("whois.internic.net", NULL);
		server = query_crsnic(sockfd, wee);
		closeconn(sockfd);
		if (!server)
			return 0;
		printf(("\nFound InterNIC referral to %s.\n\n"), server);
	}

	if (!server) {
		server = whichwhois(wee);
        switch (server[0]) {
            case 0:
                if (!(server = getenv("WHOIS_SERVER")))
                    server = "whois.internic.net";
                break;
            case 1:
                puts(("This TLD has no whois server, but you can access the "
                            "whois database at"));
            case 2:
                puts(server + 1);
                return 0;
            case 3:
                puts(("This TLD has no whois server."));
                return 0;
	}
    }

    p = queryformat(server, fstring, wee);
    strcat(p, "\r\n");
    sockfd = openconn(server, port);
    do_query(sockfd, p);
    closeconn(sockfd);
    return 0;
}

int main(void) 
{
	int i,n,counter = 0;
	FILE *atklog;
	char dirname[1024];
	char filename[2048];
	char buf[1024];
	struct dirent **namelist;
	char *p;

#ifndef SETVBUF_REVERSE
	setvbuf(stdout,NULL,_IONBF,0);
#else
	setvbuf(stdout,_IONBF,NULL,0);
#endif
	
	printf("Content-type: text/html\n\n");
	
	printf("<HTML><HEAD><TITLE>grsecurity snortparse whois</TITLE></HEAD><BODY><CENTER>\n");
	if(!strlen(getenv("QUERY_STRING")))
		return 1;
	else {
		strncpy(qstring,getenv("QUERY_STRING"),sizeof(qstring));
		wee1 = grchr(qstring,'=',strlen(qstring));
		wee1++;
		printf("<TABLE border=2 "
        "cellspacing=0 cellpadding=3 bordercolorlight=\"#333333\""
        "bordercolor=\"#FFFFFF\" bordercolordark=\"#000000\">"
	"<TR><TD bgcolor=\"5F7D9D\">"
		"<CENTER><B>NSLOOKUP</B></CENTER></TD></TR><TR><TD>");
		nslookup(wee1,1);
		if(strlen(domainname) > 0)
			nslookup(domainname,0);
		printf("</TD></TR></TABLE><br><br>");
		printf("<TABLE border=2 "
        "cellspacing=0 cellpadding=3 bordercolorlight=\"#333333\""
        "bordercolor=\"#FFFFFF\" bordercolordark=\"#000000\">"
	"<TR><TD bgcolor=\"5F7D9D\">"
		"<CENTER><B>IP WHOIS</B></CENTER></TD></TR><TR><TD><PRE>");
		whois(wee1);
		printf("</PRE></TD></TR></TABLE><br><br>");
		if(strlen(domainname) > 0){
			for(i=strlen(domainname);i>0;i--){
				if(domainname[i] == '.')
					counter++;
				if(counter == 2){
					i++;
					break;
				}
			}
			p = domainname + i;
	printf("<TABLE border=2 "
        "cellspacing=0 cellpadding=3 bordercolorlight=\"#333333\""
        "bordercolor=\"#FFFFFF\" bordercolordark=\"#000000\">"
	"<TR><TD bgcolor=\"5F7D9D\">"
	"<CENTER><B>DOMAIN WHOIS</B></CENTER></TD></TR><TR><TD><PRE>");
			whois(p);
		printf("</PRE></TD></TR></TABLE><BR>");
		}
	}
	/* print out attack logs */
	snprintf(dirname,sizeof(dirname),"%s%s",LOGBASE "/",wee1);		
	if((n = scandir(dirname,&namelist, NULL, NULL)) > 0) {
	printf("<TABLE border=2 "
        "cellspacing=0 cellpadding=3 bordercolorlight=\"#333333\""
        "bordercolor=\"#FFFFFF\" bordercolordark=\"#000000\">"
	"<TR><TD bgcolor=\"5F7D9D\">"
	"<CENTER><B>ATTACK LOGS FROM THIS IP</B></CENTER></TD></TR><TR><TD><PRE>");

		while(n--) {
			snprintf(filename,sizeof(filename),"%s/%s",dirname,namelist[n]->d_name);
			atklog = fopen(filename,"r");
			while(fgets(buf,sizeof(buf),atklog)) printf("%s",buf);
			fclose(atklog);
		}
		free(namelist);
		printf("</PRE></TD></TR></TABLE>");
	}
    printf("</CENTER></BODY></HTML>");
    return 0;
}
